using System;
using System.Collections.Generic;
using System.Text;
using Emgu.CV;
using System.Threading;
using Emgu.CV.Structure;
using System.Drawing;

namespace MotionTracking
{

    #region Enumerations declaration

    /// <summary>
    /// Width enumeration in pixels
    /// </summary>
    public enum ResolutionWidth : int
    {
        /// <summary>320 pixels</summary>
        _320 = 320,
        /// <summary>640 pixels</summary>
        _640 = 640,
        /// <summary>800 pixels</summary>
        _800 = 800,
        /// <summary>1024 pixels</summary>
        _1024 = 1024,
        /// <summary>1280 pixels</summary>
        _1280 = 1280
    }

    /// <summary>
    /// Height enumeration in pixels
    /// </summary>
    public enum ResolutionHeight : int
    {
        /// <summary>240 pixels</summary>
        _240 = 240,
        /// <summary>480 pixels</summary>
        _480 = 480,
        /// <summary>600 pixels</summary>
        _600 = 600,
        /// <summary>768 pixels</summary>
        _768 = 768,
        /// <summary>800 pixels</summary>
        _800 = 800
    }

    #region Structures declaration

    /// <summary>
    /// Point structure for 3D position.
    /// </summary>
    public struct Point3D
    {
        /// <summary>Abscissa value.</summary>
        public int X;
        /// <summary>Ordinate value.</summary>
        public int Y;
        /// <summary>Depth value.</summary>
        public int Z;

        /// <summary>
        /// Sets the coordinates of a "Point3D" point.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        public Point3D(int x, int y, int z)
        {
            X = x;
            Y = y;
            Z = z;
        }
    }

    /// <summary>
    /// Point structure for 3D position.
    /// </summary>
    public struct Point3DF
    {
        /// <summary>Abscissa value.</summary>
        public float X;
        /// <summary>Ordinate value.</summary>
        public float Y;
        /// <summary>Depth value.</summary>
        public float Z;

        /// <summary>
        /// Sets the coordinates of a "Point3DF" point.
        /// </summary>
        /// <param name="x"></param>
        /// <param name="y"></param>
        /// <param name="z"></param>
        public Point3DF(float x, float y, float z)
        {
            X = x;
            Y = y;
            Z = z;
        }
    }

    #endregion

    #endregion

    /// <summary>
    /// Infrared diodes tracking class.
    /// </summary>
    /// <remarks>To use the Camera class, <a href ="http://opencv.willowgarage.com/wiki/InstallGuide">OpenCV1.1 pre</a> and 
    /// <a href = "http://www.microsoft.com/downloads/details.aspx?familyid=200B2FD9-AE1A-4A14-984D-389C36F85647&amp;displaylang=en">MSVCRT 8.0 SP1</a>
    /// need to be installed.
    /// To enhance the infrared detection, disable the auto-mode of the Exposure function of the camera and reduce the exposure time
    /// of the camera. (5% is usually a good value)
    /// A tool like VidCap (VIDCAP32.EXE) can be used to change the settings of the webcamera.</remarks>
      /// <example>
    /// This example shows how to suscribe to <see cref="CameraFingerChanged"/> event to retrieve the position of the fingers diodes.
    /// <code>
    /// private Camera fingersTracking;
    /// 
    /// public void InitializeCamera()
    /// {
    ///     //Camera initialization
    ///     //Do not forget to disable the automatic mode of the exposure option of the webcam with a tool like VidCap (Vidcap32.exe)
    ///     fingersTracking = new Camera(ResolutionWidth._640, ResolutionHeight._480, ResolutionWidth._1024, ResolutionHeight._768);
    ///     fingersTracking.Connect();
    ///     fingersTracking.CameraFingerChanged += new Camera.CameraFingerChangedHandler(OnFingersPositionUpdated)
    ///     fingersTracking.TrackInfraredDiodes();
    /// }
    /// 
    /// private void OnFingersPositionUpdated(object sender, OnCameraFingerChangedEventArgs e)
    /// {
    /// }
    /// </code>
    /// </example>
    public class Camera : IDisposable
    {
        #region Variables declaration
        private Capture camera;
        private Boolean connected;
        private Boolean headTrackingState;
        private Thread thread;
        private List<Point> diodesPositionList;
        private List<PointF> framesPositionList;
        private int diode1CptLost;
        private int diode2CptLost;
        private int framesDiodesCptLost;
        private ResolutionWidth cameraFrameWidth;
        private ResolutionHeight cameraFrameHeight;
        private ResolutionWidth screenWidth;
        private ResolutionHeight screenHeight;
        private double minWidthBetweenDiodes;
        private double minHeightBetweenDiodes;
        private int nbFramesLostDiode = 8;
        private Boolean inverseXAxis;
        private Boolean inverseYAxis;
        private Boolean disposed;
        #endregion

        #region Enumerations declaration

        #endregion

        #region Delegates & events declaration

        /// <summary>
        /// Camera finger(s) position changed event handler. 
        /// <remarks>ONLY USED FOR DEBUGGING.</remarks>
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFingerChangedDebugHandler(object sender, OnCameraFingerChangedDebugEventArgs e);

        /// <summary>
        /// Camera finger(s) position changed event.
        /// <remarks>ONLY USED FOR DEBUGGING.</remarks>
        /// </summary>
        public event CameraFingerChangedDebugHandler CameraFingerDebugChanged;

        /// <summary>
        /// Finger(s) position changed event handler. 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFingerChangedHandler(object sender, OnCameraFingerChangedEventArgs e);

        /// <summary>
        /// Finger(s) position changed event.
        /// </summary>
        public event CameraFingerChangedHandler CameraFingerChanged;

        /// <summary>
        /// Camera finger(s) found event handler.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFingerFoundHandler(object sender, OnCameraFingerFoundEventArgs e);

        /// <summary>
        /// Camera finger(s) found event.
        /// </summary>
        public event CameraFingerFoundHandler CameraFingerFound;

        /// <summary>
        /// Camera finger(s) lost event handler.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFingerLostHandler(object sender, OnCameraFingerLostEventArgs e);

        /// <summary>
        /// Camera finger(s) lost event. 
        /// </summary>
        public event CameraFingerLostHandler CameraFingerLost;

        /// <summary>
        /// Camera Frames position changed event handler.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFramesChangedHandler(object sender, OnCameraFramesChangedEventArgs e);

        /// <summary>
        /// Camera Frames position changed event.
        /// </summary>
        public event CameraFramesChangedHandler CameraFramesChanged;

        /// <summary>
        /// Camera frames lost event handler.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFramesLostHandler(object sender, OnCameraFramesLostEventArgs e);

        /// <summary>
        /// Camera frames lost event.
        /// </summary>
        public event CameraFramesLostHandler CameraFramesLost;

        /// <summary>
        /// Camera frames found event handler.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        public delegate void CameraFramesFoundHandler(object sender, OnCameraFramesFoundEventArgs e);

        /// <summary>
        /// Camera frames found event.
        /// </summary>
        public event CameraFramesFoundHandler CameraFramesFound;

        #endregion

        #region Methods declaration

        /// <summary>
        /// Constructor 
        /// <remarks>Sets the camera frames and the screen sizes used to display the infrared diodes position.</remarks>
        /// </summary>
        /// <param name="cameraWidthTemp"></param>
        /// <param name="cameraHeightTemp"></param>
        /// <param name="screenWidthTemp"></param>
        /// <param name="screenHeightTemp"></param>
        public Camera(ResolutionWidth cameraWidthTemp, ResolutionHeight cameraHeightTemp, ResolutionWidth screenWidthTemp, ResolutionHeight screenHeightTemp)
        {
            cameraFrameWidth = cameraWidthTemp;
            cameraFrameHeight = cameraHeightTemp;
            screenWidth = screenWidthTemp;
            screenHeight = screenHeightTemp;
            minWidthBetweenDiodes = 0.09375;
            minHeightBetweenDiodes = 0.05208;
            headTrackingState = true;
            inverseXAxis = false;
            inverseYAxis = false;
            disposed = false;

            diodesPositionList = new List<Point>();
            framesPositionList = new List<PointF>();
            diodesPositionList.Add(new Point(0, 0));
            diodesPositionList.Add(new Point(0, 0));
        }

        /// <summary>
        /// Constructor 
        /// <remarks>Sets the screen size used to display the infrared diodes position.</remarks>
        /// </summary>
        /// <param name="screenWidthTemp"></param>
        /// <param name="screenHeightTemp"></param>
        public Camera(ResolutionWidth screenWidthTemp, ResolutionHeight screenHeightTemp)
        {
            cameraFrameWidth = ResolutionWidth._320;
            cameraFrameHeight = ResolutionHeight._240;
            screenWidth = screenWidthTemp;
            screenHeight = screenHeightTemp;
            minWidthBetweenDiodes = 0.09375;
            minHeightBetweenDiodes = 0.05208;
            headTrackingState = true;
            inverseXAxis = false;
            inverseYAxis = false;
            disposed = false;

            diodesPositionList = new List<Point>();
            framesPositionList = new List<PointF>();
            diodesPositionList.Add(new Point(0, 0));
            diodesPositionList.Add(new Point(0, 0));
        }

        /// <summary>
        /// Constructor.
        /// </summary>
        public Camera()
        {
            cameraFrameWidth = ResolutionWidth._320;
            cameraFrameHeight = ResolutionHeight._240;
            screenWidth = ResolutionWidth._800;
            screenHeight = ResolutionHeight._600;
            minWidthBetweenDiodes = 0.09375;
            minHeightBetweenDiodes = 0.05208;
            headTrackingState = true;
            inverseXAxis = false;
            inverseYAxis = false;
            disposed = false;

            diodesPositionList = new List<Point>();
            framesPositionList = new List<PointF>();
            diodesPositionList.Add(new Point(0, 0));
            diodesPositionList.Add(new Point(0, 0));
        }

        /// <summary>
        /// Destructor
        /// </summary>
        /// <remarks></remarks>
        ~Camera()
        {
            Dispose(false);
        }

        /// <summary>
        /// Releases all resources used by the current <see cref="MotionTracking.Camera">Camera</see> object.
        /// </summary>
        public void Dispose()
        {
            //Call GC.SupressFinalize to take this object off the finalization queue
            //and prevent finalization code for this object from executing a second time.
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Releases all resources used by the current <see cref="MotionTracking.Camera">Camera</see> object.
        /// </summary>
        /// <remarks>If disposing equals true, managed and unmanaged resources are disposed.
        /// If disposing equals false, only unmanaged resources are disposed.</remarks>
        /// <param name="disposing"></param>
        private void Dispose(bool disposing)
        {
            // Check to see if Dispose has already been called.
            if (!this.disposed)
            {
                //Close the camera thread
                if (thread.IsAlive) Disconnect();

                // If disposing equals true, dispose all managed
                // and unmanaged resources.
                if (disposing && camera != null)
                {
                    // Dispose managed resources.
                    camera.Dispose();
                }

                // Call the appropriate methods to clean up
                // unmanaged resources here.
                // If disposing is false,
                // only the following code is executed.

                // Note disposing has been done.
                disposed = true;
            }

        }

        /// <summary>
        /// Raises a OnCameraFingerChanged event : ONLY USED FOR DEBUGGING.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFingerChangedEvent(OnCameraFingerChangedDebugEventArgs e)
        {
            if (CameraFingerDebugChanged != null)
            {
                lock (CameraFingerDebugChanged)
                {
                    CameraFingerDebugChanged(this, e);
                }
            }
        }

        /// <summary>
        /// Raises a OnCameraFingerChanged event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFingerChangedEvent(OnCameraFingerChangedEventArgs e)
        {
            if (CameraFingerChanged != null)
            {
                lock (CameraFingerChanged)
                {
                    CameraFingerChanged(this, e);
                }
            }
        }

        /// <summary>
        /// Raises a OnCameraFingerLost event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFingerLostEvent(OnCameraFingerLostEventArgs e)
        {
            if (CameraFingerLost != null)
            {
                lock (CameraFingerLost)
                {
                    CameraFingerLost(this, e);
                }
            }
        }

        /// <summary>
        /// Raises a OnCameraFingerFound event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFingerFoundEvent(OnCameraFingerFoundEventArgs e)
        {
            if (CameraFingerFound != null)
            {
                lock (CameraFingerFound)
                {
                    CameraFingerFound(this, e);
                }
            }
        }

        /// <summary>
        /// Raises a CameraFramesChange event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFramesChangedEvent(OnCameraFramesChangedEventArgs e)
        {
            if (CameraFramesChanged != null)
            {
                lock (CameraFramesChanged)
                {
                    CameraFramesChanged(this, e);
                }
            }
        }

        /// <summary>
        /// Raises a CameraFramesLost event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFramesFoundEvent(OnCameraFramesFoundEventArgs e)
        {
            if (CameraFramesFound != null)
            {
                lock (CameraFramesFound)
                {
                    CameraFramesFound(this, e);
                }
            }
        }

        /// <summary>
        /// Raises a CameraFramesFound event.
        /// </summary>
        /// <param name="e"></param>
        protected virtual void RaiseCameraFramesLostEvent(OnCameraFramesLostEventArgs e)
        {
            if (CameraFramesLost != null)
            {
                lock (CameraFramesLost)
                {
                    CameraFramesLost(this, e);
                }
            }
        }

        /// <summary>
        /// Connect to the webcam.
        /// </summary>
        public void Connect()
        {
            if (connected == false)
            {
                try
                {
                    camera = new Capture();
                    camera.SetCaptureProperty(Emgu.CV.CvEnum.CAP_PROP.CV_CAP_PROP_FRAME_WIDTH, (double)cameraFrameWidth);
                    camera.SetCaptureProperty(Emgu.CV.CvEnum.CAP_PROP.CV_CAP_PROP_FRAME_HEIGHT, (double)cameraFrameHeight);
                    connected = true;
                }
                catch (Exception e)
                {
                    System.Diagnostics.Trace.WriteLine(e.Message);
                }
            }
        }

        /// <summary>
        /// Enables the fingers and head trackings.
        /// </summary>
        public void TrackInfraredDiodes()
        {
            thread = new Thread(new ThreadStart(ProcessImage));
            thread.Start();
        }

        /// <summary>
        /// Processes the image retrieved from the webcam to get the fingers position.
        /// </summary>
        private void ProcessImage()
        {

            while (connected == true && camera != null)
            {
                int i = 0, j = 0, indexDiode1 = -1, indexDiode2 = -1;
                SortedList<double, Rectangle> contoursPositionListTemp = new SortedList<double, Rectangle>();
                //Debug : show the points tracked on the image
                Image<Bgr, Byte> pictureDots = new Image<Bgr, Byte>((int)cameraFrameWidth, (int)cameraFrameHeight);

                Image<Gray, Byte> pictureCaptured = camera.QueryGrayFrame();
                if (inverseYAxis == false) pictureCaptured._Flip(Emgu.CV.CvEnum.FLIP.VERTICAL);
                if (!inverseXAxis == false) pictureCaptured._Flip(Emgu.CV.CvEnum.FLIP.HORIZONTAL);
                pictureCaptured = pictureCaptured.ThresholdBinary(new Gray(150), new Gray(255));
                pictureCaptured.Erode(1);
                pictureCaptured.Dilate(3);

                Gray cannyThreshold = new Gray(180);
                Gray cannyThresholdLinking = new Gray(120);
                Gray circleAccumulatorThreshold = new Gray(120);
                Image<Gray, byte> canny = pictureCaptured.Canny(cannyThreshold, cannyThresholdLinking);
                Contour<Point> contours = canny.FindContours();
                contours = canny.FindContours(Emgu.CV.CvEnum.CHAIN_APPROX_METHOD.CV_LINK_RUNS, Emgu.CV.CvEnum.RETR_TYPE.CV_RETR_TREE);

                if (contours != null)
                {

                    #region Contours processing

                    using (MemStorage storage = new MemStorage()) //allocate storage for contour approximation
                        do
                        {
                            if (contours.Area < 200 && contours.Area > 10 && ((float)contours.BoundingRectangle.Width / (float)contours.BoundingRectangle.Height > 0.333
     && (float)contours.BoundingRectangle.Width / (float)contours.BoundingRectangle.Height < 3))
                            {
                                if (!contoursPositionListTemp.ContainsKey(contours.BoundingRectangle.Left /
                                 (10 * (double)cameraFrameWidth) + contours.BoundingRectangle.Bottom))
                                    contoursPositionListTemp.Add(contours.BoundingRectangle.Bottom / (10 * (double)cameraFrameHeight) +
                                    contours.BoundingRectangle.Left, contours.BoundingRectangle);
                            }
                            contours = contours.HNext;
                            //   if(contours.
                        }
                        while (contours != null);
                    #endregion

                    #region Clean contours list
                    //If two contours are close to each other
                    //We process a mean of the value of the two contours and put it in a list
                    SortedList<double, Rectangle> diodesPositionListTemp = new SortedList<double, Rectangle>();


                    for (i = 0; i < contoursPositionListTemp.Count; i++)
                    {
                        for (j = i + 1; j < contoursPositionListTemp.Count; j++)
                        {
                            if (Math.Abs(contoursPositionListTemp.Values[i].Bottom - contoursPositionListTemp.Values[j].Bottom) < 10
                                                         && Math.Abs(contoursPositionListTemp.Values[i].Left - contoursPositionListTemp.Values[j].Left) < 10)
                                contoursPositionListTemp.RemoveAt(j);

                        }

                    }

                    #endregion

                    #region Search if the frames are on the picture
                    //If the frames tracking function has been enabled
                    //At least two contours has to be found, after finding them we remove them from the contours list
                    //We select the points for which the height and the width differences are below the values fixed in percent
                    //by the user
                    //  minWidthBetweenDiodes  = 0.09375 (25px) and minHeightBetweenDiodes = 0.05208 (60px) are good values for a 640*480 webcamera
                    framesPositionList.Clear();
                    List<double> frameContoursList = frameContoursList = new List<double>(); ;
                    if (headTrackingState == true && contoursPositionListTemp.Count > 1)
                    {
                        List<double> ContoursToRemoveList = new List<double>();
                        //For each diodes process the height and the width differences
                        for (i = 0; i < contoursPositionListTemp.Count; i++)
                        {
                            for (j = 0; j < contoursPositionListTemp.Count; j++)
                            {
                                //frames detection
                                if (Math.Abs(contoursPositionListTemp.Values[i].Bottom - contoursPositionListTemp.Values[j].Bottom) < minHeightBetweenDiodes * (double)cameraFrameHeight
                                    && Math.Abs(contoursPositionListTemp.Values[i].Left - contoursPositionListTemp.Values[j].Left) < minWidthBetweenDiodes * (double)cameraFrameWidth
                                    && Math.Abs(contoursPositionListTemp.Values[i].Left - contoursPositionListTemp.Values[j].Left) > 10)
                                {
                                    if (!frameContoursList.Contains(contoursPositionListTemp.Keys[i]))
                                    {
                                        frameContoursList.Add(contoursPositionListTemp.Keys[i]);
                                        framesPositionList.Add(new PointF((float)((-(double)(contoursPositionListTemp.Values[i].Left + contoursPositionListTemp.Values[i].Width) + (double)cameraFrameWidth) / (double)cameraFrameWidth * (double)screenWidth),
                                      (float)((double)(contoursPositionListTemp.Values[i].Bottom + contoursPositionListTemp.Values[i].Height / 2) / (double)cameraFrameHeight * (double)screenHeight)));
                                        //DEBUG - USED TO SHOW THE FRAMES ON THE PICTURE          pictureDots.Draw(contoursPositionListTemp.Values[i], new Bgr(255, 255, 255), 5);
                                    }
                                }
                            }
                        }
                    }
                    //If two frames diodes have been found, send an CameraFramesChanged event with the position of the frames.
                    //Remark : if it is the first time that the frame are discovered : send a OnCameraFramesFound event.
                    //Process nbFramesLostDiode images before sending a frames lost event.
                    if (framesPositionList.Count == 2)
                    {
                        RaiseCameraFramesChangedEvent(new OnCameraFramesChangedEventArgs(framesPositionList));
                        if (framesDiodesCptLost == nbFramesLostDiode) RaiseCameraFramesFoundEvent(new OnCameraFramesFoundEventArgs());
                        framesDiodesCptLost = 0;
                    }
                    else if (framesDiodesCptLost < nbFramesLostDiode - 1) framesDiodesCptLost++;
                    else if (framesDiodesCptLost == nbFramesLostDiode - 1)
                    {
                        RaiseCameraFramesLostEvent(new OnCameraFramesLostEventArgs());
                        framesDiodesCptLost++;
                    }

                    //Remove the frames contours from the contours list
                    for (i = 0; i < frameContoursList.Count; i++) contoursPositionListTemp.Remove(frameContoursList[i]);

                    #endregion

                    #region Search if one or more finger(s) point(s) the screen
                    if (framesDiodesCptLost == 0 || framesDiodesCptLost == nbFramesLostDiode)
                    {
                        //Select only the points with the most distant widths
                        //Adapt the values to the display screen size

                        //Try to match the diode seen on the screen only if two significant diodes have been found on the picture
                        if (contoursPositionListTemp.Count == 2)
                        {
                            //Diode 2 => id = 0
                            if (diode1CptLost < nbFramesLostDiode && diode2CptLost < nbFramesLostDiode
                              && Math.Pow(contoursPositionListTemp.Values[0].Left + contoursPositionListTemp.Values[0].Width / 2 - diodesPositionList[0].X, 2) + Math.Pow(contoursPositionListTemp.Values[0].Bottom + contoursPositionListTemp.Values[0].Height / 2 - diodesPositionList[0].Y, 2)
                              > Math.Pow(contoursPositionListTemp.Values[0].Left + contoursPositionListTemp.Values[0].Width / 2 - diodesPositionList[1].X, 2) + Math.Pow(contoursPositionListTemp.Values[0].Bottom + contoursPositionListTemp.Values[0].Height / 2 - diodesPositionList[1].Y, 2))
                            {
                                indexDiode1 = 1;
                                indexDiode2 = 0;
                            }
                            else if ((diode2CptLost == nbFramesLostDiode || diode1CptLost == nbFramesLostDiode)
                                && Math.Pow(contoursPositionListTemp.Values[0].Left + contoursPositionListTemp.Values[0].Width / 2 - diodesPositionList[0].X, 2) + Math.Pow(contoursPositionListTemp.Values[0].Bottom + contoursPositionListTemp.Values[0].Height / 2 - diodesPositionList[0].Y, 2)
                                > Math.Pow(contoursPositionListTemp.Values[1].Left + contoursPositionListTemp.Values[1].Width / 2 - diodesPositionList[0].X, 2) + Math.Pow(contoursPositionListTemp.Values[1].Bottom + contoursPositionListTemp.Values[1].Height / 2 - diodesPositionList[0].Y, 2))
                            {
                                indexDiode1 = 1;
                                indexDiode2 = 0;
                            }
                            else
                            {
                                indexDiode1 = 0;
                                indexDiode2 = 1;
                            }

                        }
                        else if (contoursPositionListTemp.Count == 1 && (diode1CptLost < nbFramesLostDiode || diode2CptLost < nbFramesLostDiode)
                   && Math.Pow(contoursPositionListTemp.Values[0].Left + contoursPositionListTemp.Values[0].Width / 2 - diodesPositionList[0].X, 2) + Math.Pow(contoursPositionListTemp.Values[0].Bottom + contoursPositionListTemp.Values[0].Height / 2 - diodesPositionList[0].Y, 2)
                              > Math.Pow(contoursPositionListTemp.Values[0].Left + contoursPositionListTemp.Values[0].Width / 2 - diodesPositionList[1].X, 2) + Math.Pow(contoursPositionListTemp.Values[0].Bottom + contoursPositionListTemp.Values[0].Height / 2 - diodesPositionList[1].Y, 2))
                        {
                            indexDiode2 = 0;

                        }
                        else if (contoursPositionListTemp.Count == 1 && (diode1CptLost < nbFramesLostDiode || diode2CptLost < nbFramesLostDiode))
                        {
                            indexDiode1 = 0;

                        }
                        else if (contoursPositionListTemp.Count == 1 && diode1CptLost == nbFramesLostDiode && diode2CptLost == nbFramesLostDiode)
                        {
                            indexDiode1 = 0;
                        }
                    }
                    #endregion
                }

                //Diode 1
                if (indexDiode1 > -1)
                {
                    diodesPositionList[0] = new Point((int)(Math.Abs(contoursPositionListTemp.Values[indexDiode1].Left +
                       (double)contoursPositionListTemp.Values[indexDiode1].Width / 2)),
                         (int)(Math.Abs(contoursPositionListTemp.Values[indexDiode1].Bottom +
                         (double)contoursPositionListTemp.Values[indexDiode1].Height / 2)));

                    //If it is the first time that the diode is detected : send a found event
                    if (diode1CptLost == nbFramesLostDiode) RaiseCameraFingerFoundEvent(new OnCameraFingerFoundEventArgs(1, diodesPositionList[0]));

                    //If at least one diode has been found : reset the diode lost counter
                    if (diode1CptLost > 0) diode1CptLost = 0;

                    //Debug : show the points tracked on the image
                    //     pictureDots.Draw(contoursPositionListTemp.Values[indexDiode1], new Bgr(0, 0, 255), 5);
                }
                //Increase a counter that counts the number of frames without finding a one diode (MAX_VALUE = 7 => 0.5s)
                else if (diode1CptLost < nbFramesLostDiode - 1) diode1CptLost++;

               //If no diodes has been found on the last 7 captured frames
                //Erase the 1st diode position value saved  and lauch an "lost" event
                else if (diode1CptLost == nbFramesLostDiode - 1)
                {
                    RaiseCameraFingerLostEvent(new OnCameraFingerLostEventArgs(1, diodesPositionList[0]));
                    diodesPositionList[0] = new Point((int)cameraFrameWidth, (int)cameraFrameHeight);
                    diode1CptLost++;
                }

                //Diode 2
                if (indexDiode2 > -1)
                {
                    diodesPositionList[1] = new Point((int)(Math.Abs(contoursPositionListTemp.Values[indexDiode2].Left +
                   (double)contoursPositionListTemp.Values[indexDiode2].Width / 2)),
                     (int)(Math.Abs(contoursPositionListTemp.Values[indexDiode2].Bottom +
                     (double)contoursPositionListTemp.Values[indexDiode2].Height / 2)));

                    //if it is the first time that the diode is detected : send a found event
                    if (diode2CptLost == nbFramesLostDiode) RaiseCameraFingerFoundEvent(new OnCameraFingerFoundEventArgs(2, diodesPositionList[1]));

                    //if at least two diodes has beem found : reset the diode lost counter
                    if (diode2CptLost > 0) diode2CptLost = 0;

                    //Debug : show the points tracked on the image
                    pictureDots.Draw(contoursPositionListTemp.Values[indexDiode2], new Bgr(0, 255, 0), 5);
                }
                //Increase a counter that counts the number of frames captured without finding two diodes (MAX_VALUE = 7 => 0.5s)
                else if (diode2CptLost < nbFramesLostDiode - 1) diode2CptLost++;

                       //If less than two diodes has been found on the 7 last captured frames
                //Erase the 2nd diode position value saved and launch an "lost" event
                else if (diode2CptLost == nbFramesLostDiode - 1)
                {
                    RaiseCameraFingerLostEvent(new OnCameraFingerLostEventArgs(2, diodesPositionList[1]));
                    diodesPositionList[1] = new Point((int)cameraFrameWidth, (int)cameraFrameHeight);
                    diode2CptLost++;
                }

                //Create a list of the leds position sent with the event
                List<Point> diodesPositionListEvent = new List<Point>();
                for (i = 0; i < diodesPositionList.Count; i++)
                {
                    diodesPositionListEvent.Add(new Point((int)((double)screenWidth - diodesPositionList[i].X / (double)cameraFrameWidth * (double)screenWidth),
                  (int)((double)screenHeight - diodesPositionList[i].Y / (double)CameraFrameHeight * (double)screenHeight)));
                }
                //      RaiseCameraChangedEvent(new OnCameraChangedDebugEventArgs(pictureDots, diodesPositionListEvent));
                if (diode1CptLost < nbFramesLostDiode || diode2CptLost < nbFramesLostDiode) RaiseCameraFingerChangedEvent(new OnCameraFingerChangedEventArgs(diodesPositionListEvent));
                Thread.Sleep(41);
            }


        }

        /// <summary>
        /// Disconnects from the webcam.
        /// </summary>
        /// <returns></returns>
        public Boolean Disconnect()
        {
            try
            {
                if (camera != null)
                {
                    connected = false;
                    Thread.CurrentThread.Join(1000);
                }
                return true;
            }
            catch (Exception e)
            {
                System.Diagnostics.Trace.WriteLine(e.Message);
                return false;
            }
        }

        #endregion

        #region Properties declaration

        /// <summary>
        /// Gets or Sets the X axis coordinates reverser state.
        /// </summary>
        public Boolean InverseXAxisState
        {
            get { return inverseXAxis; }
            set { inverseXAxis = value; }
        }

        /// <summary>
        /// Gets or Sets the Y axis coordinates reverser state.
        /// </summary>
        public Boolean InverseYAxisState
        {
            get { return inverseYAxis; }
            set { inverseYAxis = true; }
        }

        /// <summary>
        /// Gets or Sets the width of the frames captured by the camera.
        /// </summary>
        public ResolutionWidth CameraFrameWidth
        {
            get { return cameraFrameWidth; }
            set
            {
                cameraFrameWidth = value;
                if (camera != null) camera.SetCaptureProperty(Emgu.CV.CvEnum.CAP_PROP.CV_CAP_PROP_FRAME_WIDTH, (double)cameraFrameWidth);
            }
        }

        /// <summary>
        /// Gets or Sets the height of the frames captured by the camera.
        /// </summary>
        public ResolutionHeight CameraFrameHeight
        {
            get { return cameraFrameHeight; }
            set
            {
                cameraFrameHeight = value;
                if (camera != null) camera.SetCaptureProperty(Emgu.CV.CvEnum.CAP_PROP.CV_CAP_PROP_FRAME_HEIGHT, (double)cameraFrameHeight);
            }
        }

        /// <summary>
        /// Gets or Sets the width of the screen used to display the fingers position.
        /// </summary>
        public ResolutionWidth ScreenWidth
        {
            get { return screenWidth; }
            set { screenWidth = value; }
        }

        /// <summary>
        /// Gets or Sets the height of the screen used to display the fingers position.
        /// </summary>
        public ResolutionHeight ScreenHeight
        {
            get { return screenHeight; }
            set { screenHeight = value; }
        }

        /// <summary>
        /// Gets or Sets the minimum width between two significant diodes.
        /// <para>(in percent of the camera frames width)</para>
        /// <remarks>Two diodes are insignificant only if their height
        /// and width differences are below the fixed limit.</remarks>
        /// </summary>
        public double MinWidthBetweenDiodes
        {
            get { return minWidthBetweenDiodes * 100; }
            set
            {
                if (value > 0 && value < 100) minWidthBetweenDiodes = value / 100;
                else if (value < 1) minWidthBetweenDiodes = 0.01;
                else minWidthBetweenDiodes = 0.99;
            }
        }

        /// <summary>
        /// Gets or Sets the minimum height between two significant diodes.
        /// <para>(in percent of the camera frames height)</para>
        /// <remarks>Two diodes are insignificant only if their height
        /// and width differences are below the fixed limit.</remarks>
        /// </summary>
        public double MinHeightBetweenDiodes
        {
            get { return minHeightBetweenDiodes * 100; }
            set
            {
                if (value > 0 && value < 100) minHeightBetweenDiodes = value / 100;
                else if (value < 1) minHeightBetweenDiodes = 1;
                else minHeightBetweenDiodes = 0.99;
            }
        }

        /// <summary>
        /// Gets or Sets the maximum number of frames retrieved from the camera 
        /// between the last apparition of a diode on a frame
        /// and the loss report of the diode.
        /// </summary>
        /// <seealso cref="OnCameraFingerLostEventArgs"/>
        public int NbFramesLostDiode
        {
            get { return nbFramesLostDiode; }
            set { nbFramesLostDiode = value; }
        }

        /// <summary>
        /// Gets or Sets the state of the frames discovery function.
        /// </summary>
        public Boolean HeadTrackingState
        {
            get { return headTrackingState; }
            set { headTrackingState = value; }
        }

        #endregion
    }
}
